
# <b>Datenbank</b> regelt die Zugriffe auf die Datei "dateis" der Dateien sowie
# die Aktionen und sonstigen Informationen, die im Zusammenhang damit stehen 
#
# auftretende Fehler werden dabei idR. als Programmierfehler betrachtet: keine Ausgabe ber @notice

class Datenbank
    
  attr_reader   :data, :treelist, :view_name, :treelist_header, :notice, :info, :conn, :gone
  cattr_reader  :modul, :protocol
  
  # <b>initialize</b> liest entweder die Datenstze der "dateis"-Datei ein oder
  # die Dateien der angebundenen Datenbank von Parameter "conn" ein (nur fr MySQL)
  # Soll eine Aktualisierung durchgefhrt werden, wird berprft, ob fr jede
  # der vorgefundenen Dateien bereits ein Eintrag in der "Datei der Dateien" (dateis) existiert.
  # * Wenn nicht, wird ein Datensatz angelegt.
  # * Wenn ja, wird der Datensatz ggf. aktualisiert
  # * Liegt sogar mehr als nur ein Datensatz vor, wird eine Fehlermeldung ausgegeben.
  # Die eingelesenen Daten werden gem den Treelist-Anforderungen (#) aufbereitet
  
  def initialize(*parm)    # aufgerufen von new    
    @gone                 = 'Feld Fehlt'             # Kennzeichnung gelschter Felder
    @errors, @errors_type = [0,0,0], {'table' => []}
    
    @@dataview  = true  if !defined?(@@dataview)        # alle Verbindungen anzeigen  
    @@data      = {}    if !defined?(@@data)
      
    if parm[0] == "modul"
      @@modul = sql('DISTINCT modul','') unless @@modul
    elsif parm[0] == "noDB"
      @view_name  = "NaviSeite-0"
      datei_to_tree                       # Einlesen der datei-Datenstze
    
    else      
      @view_name  = "NaviSeite-1"  
      check_conn(parm[0])                 # @conn als Voraussetzung fr das Einlesen der physikalischen DB
      update     = parm[1]            
      
      unless @notice  
        # die Angaben sollen aktualisiert werden im Falle von "update" sowie bei einer unbekannten Verbindung
        if !@@data.key?(@conn.to_s) || update       
          files                           # @@data
        end    
        view(update)                      # @data     
        if update
          updated = do_update
          if updated > 0
            view                          # verbleibende Fehler einlesen
          end
        end      
      end 

    end   # parm[0]
  end   #initialize 
  
  # <b>change_view</b> zeigt entweder alle Dateien an
  # oder reduziert sie auf die aktuelle Verbindung @conn
       
  def change_view
    @@dataview = !@@dataview
    view
  end  
  
 ######################################################################################################   
 
  private  
   
  # <b>datei_to_tree</b> liest die Datenstze aus der Datei ein und
  # bereitet die eingelesene Datenliste fr die Treelist-Anzeige auf
  
  def datei_to_tree
    raise "Fehler in #{File.basename(__FILE__)}, #{__LINE__}: \"#{Datei.connection.adapter_name}\" beherrscht kein \"distinct\"" unless Datei.connection.supports_count_distinct?
    # Abarbeiten der hinterlegten Adapter, Datenbanken und Module
    @treelist = do_group(1)
    @@data[@view_name]  = @treelist
    treelist_head
   end
  
  # <b>check_conn</b> prft die angeforderte Connection, ob sie aktiv ist
  # oder ob die Datenbank untersttzt wird. Wird keine Connection angegeben,
  # wird als Defaultwert "Datei" vorgegeben
 
  def check_conn(conn)
  
    conn  = conn || "Datei" 
     
    if eval(conn).__id__       # existiert die Verbindung?
      @conn = ObjectSpace._id2ref(eval(conn).__id__) 
    else    
      @notice = "Die Verbindung &quot;#{conn}&quot; konnte nicht gefunden werden"
    end
    
    if @conn.connection.adapter_name != "MySQL" then
      @notice = " Datenbank &quot#{@conn.connection.adapter_name}&quot wird nicht unterst&uuml;tzt (nur MySQL)"
    elsif !@conn
      @notice = " Datenbank &quot#{@conn.to_s}&quot ist noch nicht angebunden"
    end
    
    @info = ['error', 'main'] if @notice      # Ausgabe fr die Fehlermeldung ist blicherweise das Hauptfenster

  end
   
  # <b>files</b> liest die Dateien der angebundenen Datenbank ein
  
  def files
     
      files             = {}
            
      tables            = @conn.connection.tables          # tables : definiert fr MySQL, PostgreSQL
      tables.each do |table|
        keys            = primary_key(table)
        file_full       = @conn.connection.columns(table)  # columns: definiert fr MySQL, PostgreSQL
        
        # Aufbereitung der Daten        
        file, type      = [], []
        file_full.each { |column|
          file << column.name
          type << column.type}
                
        files[table]    = keys, file, type, file_full
        
      end   #table
     
      @@data[@conn.to_s] = files      
      
  end  
  
  # <b>view</b> bereitet die Datenliste auf  
  # dabei wird die Liste alle bisher eingelesenen Daten auf die anzuzeigenden reduziert
  # (im Falle von @@dataview = false)
  # und fr diese Instanz ber <i>@data</i> zur Verfgung gestellt
  # Fehler werden nur ausgegeben, falls kein Update verlangt ist
  
  def view(update = false)
      tablefields, tablefieldsall       = {}, {}
     
      @@data.each do |conn, data|
            
        if conn.match('Navi')
          next                  # diese Routine ist nur fr die Bearbeitung der physikalischen Datei-Daten gltig
        elsif !@@dataview && @conn.to_s != conn
          next
        end

      data = data.sort
      
        # Schlsselfelder und Dateifelder befinden sich in den Arrays 0,2 der Datei-Arrays
        data.each do |table, fields|
          tablefields[table]    = fields[0,2]
          
          # Treelist-Info-Array (Sortiermerkmal, Hash... s. tree_fields): Feldnamen plus Feldinfo, aufbereitet fr treelist_helper
          sort                  = tree_fields(fields, table, update)               
          action                = ["edit_called_by_tree", "dateis"]  # Verarbeitung auf Datei-Ebene
          tablefieldsall[table] = [table, sort, [], action, [], [table, conn, @conn.connection.adapter_name] ]
                      
          end   # data
          
      
        # der Connection-Name wird in @data fr die Oberflche (application.rb) mitgegeben
        @data             = {}    if !defined?(@data)
        @data[conn] = tablefields
        
        # aufbereitete Form Treelist-Info-Array (Sortiermerkmal, Hash... s. tree_fields)
        @treelist         = {}    if !defined?(@treelist)  
        action            = ["list", "felds"]                         # Felder nach Dateien
        @treelist[conn]   = [@conn.connection.current_database, tablefieldsall, [], action, [], [@conn.connection.current_database]]  # nur MySQL, kein Datensatz vorhanden: keine id
        
      end   # @@data
      
      treelist_head
  end
  
  # <b>update</b> berprft, ob fr jede der vorgefundenen Dateien 
  # bereits ein Eintrag in der "Datei der Dateien" (dateis) existiert.
  # * Wenn nicht, wird ein Datensatz angelegt.
  # * Wenn ja, wird der Datensatz ggf. aktualisiert
  # * Liegt sogar mehr als nur ein Datensatz vor, wird eine Fehlermeldung ausgegeben.
  # Danach wird die "Datei der Felder" (felds) abgearbeitet: #do_fields
    
  def do_update

    @notice = "Datei der Dateien wurde aktualisiert<br>"
    @info   = ['hinweis', 'main']                         # Ausgabe fr die Fehlermeldung ist blicherweise das Hauptfenster    
    datei_conn  = eval("Datei")                           # Modell der Datei der Dateien  
    
    updated     = 0
    
    @data[@conn.to_s].each do |table, fields|
    
      # fr die gesuchte Datei "table" werden alle Eintrge in der Datei der Dateien bestimmt
      # falls mehr als einer vorliegt, ist das ein Fehler, wenn weniger vorliegt, muss ein Satz erstellt werden 
      files       = datei_conn.find_all_by_name(table)

      if files.length > 1
       @notice = @notice, " vorgefundene Dateien f&uuml;r den Schl&uuml;ssel &quot;#{table}&quot; = #{files.length} <i>(bitte selbst pr&uuml;fen und &auml;ndern!)</i> </br>"      
       @info   = ['error', 'main']                        # Ausgabe fr die Fehlermeldung ist blicherweise das Hauptfenster
      else
        if  files.length  == 0
          datei           = datei_conn.new
        else
          datei           = files[0]
        end
      
        attributes                  = datei.attributes
        attributes["name"]          = table
        attributes["beschreibung"]  = table if attributes["beschreibung"].empty?
        attributes["modul"]         = 'main' if attributes["modul"].empty?
        attributes["adapter"]       = @conn.connection.adapter_name         # abstract_adapter.rb: adapter_name ist gltiges Attribut aller Adapter   
        attributes["datenbank"]     = @conn.connection.current_database     # nur MySQL
        attributes["records"]       = 1000 if attributes["records"] == 0
        attributes["key"]           = fields[0].join(", ")
 
        attributes["fields"]        = Marshal_out(fields, @conn.connection.columns(table))
      
        unless attributes == datei.attributes
          datei.attributes  = attributes
          datei.save
          updated           += 1
        end
        
        updated             += do_fields(table, fields[1], datei)
        
      end   #files.length
      
    end   #table
    updated
  end
  
  ######## 2. Level    
   
  # fr Treelist: Feldnamen plus Feldinfo, aufbereitet fr treelist_helper 
  # * sort       : Sortiermerkmal
  # * Array      : standardisierte Form (Info-Array)
  #   1. Array[0]: key           - anzuzeigendes Element
  #   2. Array[1]: Hash          - enthlt den nchsten Zweig, falls vorhanden
  #   3. Array[2]: Array         - URL/Name des Icons, Alt/Tooltip zum Image
  #   4. Array[3]: Array         - Aktion und Controller
  #   5. Array[4]: Array         - Satznummer, Fehler-Icon, Fehler-Text
  #   6. Array[5]: Array         - enthlt weitere Informationen fr Tooltips, Rechte Maus
  
  def tree_fields(fields, table, update, *parm)  
    sort                  = {}          # Hash der anzuzeigenden Felder
    icon, detail          = [], []      # Arrays der weiteren Informationen    
        
    (0...fields[1].length).each do |i|
      type                = ""
      if defined?(fields[2][i])
        type  = "Feldtyp = #{fields[2][i]}"
      else
        type  = "Feldtyp = #{parm[0][i]}"
      end
      detail  = fields[1][i], "Feldnummer = #{i+1}", type           # kein Datensatz vorhanden: keine id
  
      ic      = fields[0].include?(fields[1][i]) ? "tree/key.ico" : ""
      icon    = [ic, ic.empty? ?   "" : "Schluesselfeld" ]
      # Verifizierung des Datensatzes
      id, detail[0]   = check_rec(table, fields[1][i], update, detail[0])

      sort[i.to_s]    =                             # Sortiermerkmal
         fields[1][i],                              # key
         {},                                        # leeres Hash - kein weiterer Zweig 
         icon,                                      # Array: Icon-URL, Alt/Tooltip
         ["edit_called_by_tree", "felds"],          # Aktionen auf Feldebene
         id,                                        # Satznummer, Fehler-Icon, Fehler-Text
         detail                                     # Array: Tooltips, Rechte Maus
      end   # i
    sort 
  end  
  
  # da connection.indexes gerade den Primary Key berspringt
  # und da ich keine Rails-Funktion finden kann, muss der Primrschlssel selbst bestimmt werden
  
  def primary_key(table)
    sql = "SHOW KEYS FROM #{table} WHERE key_name = 'PRIMARY'"
    keys = []
    @conn.connection.execute(sql).each { |row| keys << row[4]} # execute: definiert fr MySQL, PostgreSQL
    keys                                                       # row[4] ist Feldname
  end
 
 # <b>Marshal_out</b> bereitet die Daten fr das Dateifeld "fields" so auf,
 # dass beim Zurckspeichern die Prozedur "tree_fields" aufgerufen werden kann,
 # um die Feldinformationen identisch aufbereitet zu erhalten.
 # Dies erfolgt, da die aufbereiteten Informationen selbst nicht mehr "marshalbar" sind (zu komplex).
  
  def Marshal_out(sort, conn)
    type = []
    conn.each {|element| type << element.type}
    Marshal.dump([sort, type])
  end
  
  # <b>do_group</b> arbeitet die hinterlegten Adapter, Datenbanken und Module ab 
  
  def do_group(level, *parm)
    adapter   = parm[0] || ""
    datenbank = parm[1] || ""
    modul     = parm[2] || ""
    
    items   = {}
    
    # passende Datenstze finden
    group = find(level, adapter, datenbank, modul)
    
    # Aufbereiten der einzelnen Hierarchien
    mbr, todo, icon, action, get_id  = get_level(level)    

    group.each do |member|
      id                    = get_id.empty? ? '' : eval(get_id)
      items[member[mbr[0]]] = [member[mbr[1]], eval(todo.to_s), icon, action, [id], [member[mbr[2]]]]       # Array: s. Dokumentation "tree_fields"
      end
    items
  end
  
  # <b>find</b> ruft den SQL-Befehl auf, da der Rails-Find
  # die Anordnung der Felder zerhagelt
   
  def find(level, ad, db, md)
    values  = get_level(level) 
    where   = val_to_where(values[6], ad, db, md)
    sql(values[5],where)                                         
  end
  
  # <b>sql</b> fhrt den SQL-Befehl aus und bergibt das Ergebnis
  
  def sql(select, where)
    sql = "SELECT #{select} FROM dateis #{where}"
    group = []
    Datei.connection.execute(sql).each { |row| group << row}   # execute/distinct: definiert fr MySQL, PostgreSQL
    group  
  end
  
  # <b>val_to_where</b> erledigt die Aufbereitung der Parameter
  
  def val_to_where(where, ad, db, md)
    where   = where.gsub('\'ad\'', "\'#{ad}\'")
    where   = where.gsub('\'db\'', "\'#{db}\'") 
    where   = where.gsub('\'md\'', "\'#{md}\'") 
  end
  
  # <b>Marshal_in</b> bereitet die Feldinformationen auf
  
  def Marshal_in(stored, table)
    begin
      fields_and_types  = Marshal.load(stored)
      fields            = fields_and_types[0]  
      types             = fields_and_types[1] 
      
      tree_fields(fields, table, false, types)    # kein Update beim Einlesen der Datei "Datei"
    rescue              # leere Daten sollte heien, dass der Datei-Datensatz noch nicht mit den Informationen aus der physikalischen DB versorgt wurde
      {}                # leeres Hash - keine weitere Treelist-Verzweigung
    end    
  end

  ######## Vorgaben
     
  # <b>get_level</b> bestimmt die Hierarchie-Ebenen
  
  def get_level(level)
    get_id   = ''
    case level
    when 1
      mbr    = [0,0,0]
      todo   = "do_group(2, member[0])"
      icon   = []
      action = []                               # bisher keine Aktionen auf Ebene der Adapter
      member = "DISTINCT adapter"
      where  = ""
    when 2
      mbr    = [1,1,1]
      todo   = "do_group(3, adapter, member[1])"
      icon   = []
      action = ["list", "felds"]                         # Felder nach Dateien
      member = "DISTINCT adapter, datenbank"
      where  = "where adapter = \'ad\'" 
    when 3
      mbr    = [2,2,2]
      todo   = "do_group(4, adapter, datenbank, member[2])"
      icon   = []      
      action = ["list_modul", "dateis"]      
      member = "DISTINCT adapter, datenbank, modul"
      where  = "where adapter = \'ad\' and datenbank = \'db\'"      
    else
      mbr    = [0,1,2]
      todo   = "Marshal_in(member[9], member[mbr[1]] )"
      icon   = []      
      action = ["edit_called_by_tree", "dateis"]  # Verarbeitung auf Datei-Ebene
      member = "*"
      where  = "where adapter = \'ad\' and datenbank = \'db\' and modul = \'md\'"
      get_id = "member[0]"
    end
    return mbr, todo, icon, action, get_id, member, where  
  end
  
  # <b>treelist_head</b> bereitet die Kopfinformationen der Treelist auf
  
  def treelist_head
    @treelist_header    = []
    if @view_name == "NaviSeite-0" 
      @treelist_header[0] = "Dateiverzeichnis"                      # Root-Name
    else
      @treelist_header[0] = "Datenbankverzeichnis"                  # Root-Name    
    end
    @treelist_header[1] = ["tree/root", "Datenbanken und Dateien"]  # icon, alt/tooltip 
    @treelist_header[2] = []                                        # keine Verarbeitung auf Root-Ebene
    do_fehler                                                       # bei Bedarf Summenfehler ausgeben
  end
  
  # <b>do_fields</b> arbeitet die Dateifelder pro Datei ab:
  # * Liegt ein Dateifeld vor, bleibt es unangetastet
  # * Liegt ein Dateifeld noch nicht vor, wird es angelegt
  # * Liegt ein Dateifeld nicht mehr vor, wird es gekennzeichnet
  
  def do_fields(table, fields, file)

    updated     = 0
    
    dateis      = Feld.find_all_by_datei(table)
    dat_felder  = []
    dateis.each {|df| dat_felder << df.attributes['name'] }      unless dateis.empty?   

    # Vergleich der physikalisch existierenden Dateifelder (in "fields")
    # mit den in der Datei "felds" bereits vorhandenen (in "dat_felder")
    felder      = {}            
    fields.each do |field|
        felder[field] = dat_felder.delete(field) ? true : false
      end
    
    # "brig gebliebene" Felder in der Datei "felds" (in "dat_felder")
    # die tatschlich aber nicht mehr existieren
    dat_felder.each do |field|
      dateis.each do |df|
         if df.attributes['name'] == field && !df.attributes['beschreibung'].include?(@gone)
          df.beschreibung         += " - " + @gone + " (ehemals in Datei \"#{df.datei}\")"
          df.datei                = @gone
          df.eigenschaft          = 0
          ok                      = df.save_without_validation
          updated                 += 1      if ok  
         end
        end
      end
    
    # Erstellen der erforderlichen Datenstze
    felder.each do |key, value|
      if value  == false
        feld    = Feld.new( :name         => key,       # Beschreibung wird bei Validierung automatisch eingefgt, falls leer
                            :datei        => table)
        feld.save
        # Datensatz ist fehlerhaft, jedoch nur wegen fehlender Eigenschaft
        if feld.errors.count == 1 && !feld.errors['eigenschaft'].nil? 
          Feld.new( :name         => key,
                    :datei        => table).save_without_validation  # auch: save(false)
        end
        updated           += 1
      end         
      end     # each do    
    updated      
  end
  
  # Datensatz prfen und Fehler ausgeben, falls kein Update verlangt wird
  
  def check_rec(table, feldname, update, beschr)
            
      # Dateiangaben - werden nur fr Fehlermeldungen bentigt
      if @errors_type['table'].include?(table)
        fehler      = "Datei \"#{table}\" fehlt f&uuml;r Feld \"#{feldname}\""
      else
        @datei      = Datei.find_by_name(table)  if @datei.nil? || !( @datei.name == table)
        if @datei.nil?
          fehler    = "Datei \"#{table}\" fehlt"
          @errors_type['table'] << table
          @errors[0] += 1
        else
          feld      = Feld.find_by_name(feldname, :conditions => [ "datei = ?", table])      
          if feld.nil?
            fehler    = "Feld fehlt in Datei 'felds' f&uuml;r Feld \"#{feldname}\""
            @errors[1] += 1      
          else            
            beschr    = feld.beschreibung
            id        = [feld.id]
            # Prfung, ob Feld fehlerhaft ist - muss aktuell sein (nicht cachen)
            if feld.eigenschaft == 0
              fehler    = "Eigenschaft fehlt f&uuml;r Feld \"#{feldname}\""
              @errors[2] += 1
            end
          end
        end     # @datei
      end

    @errors[0, 2] = [0, 0]      if update   # update sollte die Probleme (auer Eigenschaft) lsen

    # Satznummer und Fehler melden
    id  = [0]  unless id
    
    id[1,2] = ['tree/faulty.ico', fehler]  if fehler                    
    return id, beschr
  end

  # Summenfehler ausgeben
  
  def do_fehler
    sum       = 0
    @errors.each {|i| sum += i }
    if sum > 0
      txt       = ''
      @errors_type['table'].each {|a| txt << a + ', ' }
      @notice   = ''
      @notice   = "#{@errors[0]}x fehlt die Datei (#{txt}) => bitte Datenbank aktualisieren"   if @errors[0] > 0
      @notice   = @notice + "<br>"  unless @notice.empty? || (@errors[1] == 0)
      @notice   = @notice + "#{@errors[1]}x fehlt die Dateibeschreibung f&uuml;r ein Feld => bitte Datenbank aktualisieren"   if @errors[1] > 0
      @notice   = @notice + "<br>"  unless @notice.empty? || (@errors[2] == 0)
      @notice   = @notice + "#{@errors[2]}x fehlt die Eigenschaft"    if @errors[2] > 0
      @info     = ['error', 'Treelist_Top', sum, @notice]     unless @notice.empty?
    end
  end
  
end     # class
