
require 'treelist'

# <b>Treelist_Modul</b> stellt die Schnittstelle zwischen Treelist-Objekten und dem Controller dar
# @default setzt dabei den Unterlassungswert fr die Treelist - ob sie anfangs auf- und zugeklappt ist 
#
# auftretende Fehler werden dabei als Programmierfehler betrachtet: keine Ausgabe ber @notice

module Treelist_Modul
  
  attr_reader   :mark_id, :parent_id, :prv_hlp
  attr_accessor :tree_html, :xxx

 # <b>tree_response</b> nimmt die Anforderungen vom Browser entgegen und verteilt sie
  
  def tree_response
    trees_and_partials(params[:html_name])  unless defined?(@@trees)
    
    tree          = @@trees.fetch(params[:html_name])  
    tree.do_html(params[:html_id])
    render :partial  => "application/treelist", :locals => { :name => params[:html_name]} 
     
  end

 # <b>treelist_show</b> ermittelt die anzuzeigende Baumstruktur

  def treelist_show(partial)    

    if @@trees.key?(partial)       
      @tree_html                  = @@trees.fetch(partial).tree_html      
    else
      raise "Fehler in #{File.basename(__FILE__)}, #{__LINE__}: \"#{partial}\" ist kein gueltiges Partial"
    end 

  end
  
  # <b>reload_tree</b> 
  
  def reload_tree
    if params[:target] == "NaviSeite-2"
      view_tree("Eigen", "", "clear")
    else
      view_tree("noDB", "", "clear")
    end
    render :partial => "application/treelist",  :controller => "application", :locals => { :name => @view_name}
  end
  
  # <b>database</b> fordert die bersicht der Datenbanken sowie ggf. ihre Aktualisierung an
      
  def database
    clear = params[:update] ? 'clear_all' : 'clear'
    
    view_tree(params[:conn], params[:update], clear)      # params[:conn]noch inaktiv    
    if params[:update]
      Feld.refresh_fields
      Eigenschaft.refresh_verwendung
    end
    render :partial => "application/nav", :controller => "application"
  end
  
  # <b>auto_complete_for_application_search</b> fordert alle passenden Treelist-Elemente an
  
  def auto_complete_for_application_search
    @items          = @@trees[@view_name].find_all(params[:application][:search])
    unless @items.empty?
      @search_list  = ['<ul>']
      @items.each { |item|  @search_list << "<li>#{item[0]}</li>"}
      @search_list  << '</ul>'
      @searchtext   = @search_list.to_s
    else
      @searchtext   = ''
    end
    render :inline =>  @searchtext
  end
  
  # <b>position</b> fordert die Suche an
  
  def position  
    if params[:input].empty?
      @info                 = ['error', 'Treelist_Top']
      flash[:notice]        = "Bitte zuerst <br>ein Element festlegen"
    else
      @mark_id, @parent_id  = @@trees[@view_name].position(params[:input])
    end
    render :partial => "application/treelist", :locals => { :name => @view_name}
  end

  # <b>do_error?</b> zeigt an, ob bei einer Treelist-Auflistung Fehler vorkamen
  
  def do_error?
    @@do_errors[@view_name]
  end

  # <b>dropped</b> empfngt das Drop-Signal und prft, ob das annehmende Element zum rufenden passt
 
  def dropped
    dropped                 = params[:dropped] ? params[:dropped] : 0    
    idnr                    = params[:idnr].scan(/[0-9]+|\w+/)    # aus :with
    view                    = params[:view]
    controller              = @@dropdrag[idnr[2]]['me']
    reply                   = controller.send(:dropped, view, dropped, idnr)  # Anschauungsbeispiel fr "send"
    @dropdrag, @html_flash  = reply[0], reply[1]

    render :partial         => 'application/dropdrag',
                            :locals => {
                            :index => idnr[1],		                    
		                    :act => 'dropped', :cont => controller.controller_name, :idnr => idnr[0]
		                    }                  
  end

  # <b>drop_text</b> ermittelt die Drop-Texte
  
  def drop_text(idnr, cont, index)
    @dropdrag       = [0]   unless @dropdrag
    @@dropdrag      = {}    unless defined?(@@dropdrag)
    
    if cont == self.controller_name
      @@dropdrag[cont]            = {}  if @@dropdrag[cont].nil?
      @@dropdrag[cont]['me']      = self            # Anschauungsbeispiel fr "send"
     end
    index = index.to_s.to_i       if index.class == String
    dropped = @dropdrag[index]  

    show    =  @@dropdrag[cont]['me'].droptext(idnr, index, dropped)
    show[4] = show[1].empty? ? '' : show[4]
    show
  end
 
 ######################################################################################################   
   
  private

  # <b>view_tree</b> bereitet das angeforderte Datenbankobjekt auf
    
  def view_tree(*parms)         
    @view_name  = Treelist.new('prv').prv

    unless parms.length == 0 && !@view_name.empty?
      parms[0]    = "noDB"              if parms.length == 0      # neuer Thread
      clear       = parms[2] || ""      
  
      if parms[0] == 'Eigen' 
        @data     = Eigenschaften.new(clear)
        @@gone    = ''                  unless defined?(@@gone)
      else
        @data     = Datenbank.new(parms[0], parms[1])      
        @database = @data.data
        @@gone    = @data.gone     
      end  

      # aktuelle Sicht     
      @view_name  = @data.view_name                  
      Treelist.new('prv', @view_name)

      # Info-Bereich
      @@do_errors   = {}                  unless defined?(@@do_errors)
      @@do_errors[@view_name]   = false
      @info                     = @data.info
      if @info
        @info[1]                = 'Treelist_Bottom'   if ['', 'main'].include?(@info[1])
        @@do_errors[@view_name] = true                unless @info[2].nil?  
      end
        
      trees_and_partials(@view_name, @data.treelist, @data.treelist_header, clear)
      flash[:notice] = @data.notice if params       # flash wird nur ausgegeben, falls @notice nicht leer ist / @params wegen Initialisierung
    end
  end
  
  # <b>trees_and_partials</b> speichert die Zuordnung der Treelist zum Partial ab
  
  def trees_and_partials(partial, treelist, treelist_header, clear)
    @default          = '-1'          # Baum soll aufgeklappt sein
    @@trees           = {}            if !defined?(@@trees) || (clear == 'clear_all') 
    @@trees.delete(partial)           if clear == 'clear'    
    unless @@trees.key?(partial)
      @@trees[partial]  = Treelist.new(partial, treelist, treelist_header, @default)
    end
  end

end     # module
