class Feld < ActiveRecord::Base

  include Models_modul
  
  # <b>refresh_fields</b> prft und ggf. korrigiert die Verwendung
  
  def Feld.refresh_fields
    all         = find(:all)
    all.each do |feld|
        feld.valid?                       # falsche Eigenschaften, Redundanzen, Vorgnger, Nachfolger werden eliminiert
        feld.save_without_validation      
      end
  end
  
  # <b>save_me</b> fhrt das Speichern durch - Validierung erfolgte zuvor
  
  def save_me(prv)
    if save_without_validation
      if prv[0] != self.eigenschaft 
        @eigen.save(prv[0])
      end
      if prv != [self.eigenschaft, self.redundant, self.vorgaenger, self.nachfolger]
       Feld.refresh_fields
      end
    end
  end

  # <b>exit?</b> prft darauf, ob das Feld ausgegeben wird
  
  def exit?
    self.exitdistanz > self.anzahl
  end
 
  ######################################################################################################    
  
  private  
 
  # <b>calc_conseq</b arbeitet den gesamten Datensatz ab und bestimmt rechnerische Konsequenzen
  # ! Voraussetzung ist, dass die validate-Feld-Prfungen erfolgt sind
 
  def calc_conseq(attr, store)

    if store
      # bei korrekter Redundanz: nur id des charakteristischen Datenfeldes ist magebend
      if @master      
        nochange    = ['id', 'name', 'datei', 'beschreibung', 'redundant']          
        keys        = attr.keys - nochange
        keys.each {|key| attr[key] = @master.attributes[key]}
      else
        
        # Berechnungen von Distanzen und Gewichten
        anzahl_Dateien
        attr['eigengewicht']                          = eigengew(attr['anzahl'])
        attr['tentakelzahl'], attr['profilgewicht']   = tentakel(@eigen.id)
        attr['portaldistanz'], attr['portalgewicht']  = distanz('vorgaenger') 
        attr['exitdistanz'], attr['exitgewicht']      = distanz('nachfolger')
        
        attr['typfeld'], attr['typaufgabe']           = typ(attr)
      end
      
      attr
    end
    
  end     # calc_conseq
  
  # Prfung der einzelnen Felder
  
  def datei_ok?(check)            # alle Felder sind als Methoden bereits angelegt: Der Call darauf gibt den Feldwert zurck
    @datei = Datei.find_by_name(check)      # Datei muss einfach existieren: auer im Fall von "Leichen"
    @errors.add("datei", "Die Datei \"#{check}\" kann nicht gefunden werden") unless @datei
    @datei ? @datei.name : check
  end
  
  def beschreibung_ok?(check)
    check = self.datei.capitalize + '_' + self.name.capitalize    if check.empty?  
    check
  end
    
  def redundant_ok?(check)
    related(check, 'red')
    if @master &&  self.datei != @master.datei
      @errors.add("redundant",
            "Das Feld \"#{self.name}\" befindet sich nicht in der Datei \"#{@master.datei}\"
             des als redundant angegebenen Feldes \"#{@master.name}\".")
      @master = nil
    end
    @master ? @master.id : 0
  end
  
  def anzahl_ok?(check)
    if check.to_i == 0 && @datei
      check = @datei.records
    elsif check.to_i == 0
      check = 100
    end
    check
  end

  def eigenschaft_ok?(check)
    begin
    @eigen = Eigenschaft.find(check.to_i)   if check.to_i > 0  
      unless @eigen
        @eigen = Eigenschaft.find_by_name(self.name)
        unless @eigen
          eigen = Eigenschaft.find(:all, :conditions => "name like \'%#{self.name}%\' ")
          if eigen.length == 1
            @eigen = eigen[0]
          elsif eigen.length > 1
            cand = eigen.select { |rec| rec.attributes['name'].include?(@datei.name) }
            cand = eigen.select { |rec| rec.attributes['name'].include?(@datei.name[0,4]) } if cand.length == 0
            cand = eigen.select { |rec| rec.attributes['name'].upcase.include?(@datei.name[0,4].upcase) } if cand.length == 0
            @eigen = cand[0]  if cand.length == 1
          end
  
        end     # unless #2        
      end
    rescue
    end
      
    # immer noch nichts gefunden
    msg = check.to_i > 0 ? "Die Eigenschaft \"#{check}\" existiert nicht" : "Es liegt keine Eigenschaft vor - sie wird aber gebraucht"
    @errors.add("eigenschaft", msg)  unless @eigen
    
    @eigen ? @eigen.id : 0    
  end  
  
  def vorgaenger_ok?(check)
    related(check, 'vor')
    if @vor &&  self.datei == @vor.datei
      @errors.add("vorgaenger",
            "Das Feld \"#{self.name}\" befindet sich in derselben Datei \"#{@vor.datei}\"
             wie das  als Vorg&auml;nger angegebene Feld \"#{@vor.name}\".")
      @vor = nil
    end
    @vor ? @vor.id : 0
  end
  
  def nachfolger_ok?(check)
    if check.to_i < 0      # Endstation - keine Ausgabe vorgesehen
      @nach = Feld.new(:name => "Endstation", :nachfolger => 0)
      @nach.id = -1
    else
      related(check, 'nach')
      if @nach &&  self.datei == @nach.datei
        @errors.add("nachfolger",
              "Das Feld \"#{self.name}\" befindet sich in derselben Datei \"#{@nach.datei}\"
               wie das als Nachfolger angegebene Feld \"#{@nach.name}\".")
        @nach = nil
      end
    end
    @nach ? @nach.id : 0
  end
  
  # Berechnungen von Gewichten und Distanzen
  
  def anzahl_Dateien
    @p = self.connection.tables.length      # fr das gesamte Datenbanksystem
    @p = 1      if @p == 0
  end
  
  def eigengew(check)
    check > 1 ? 1.0/(check-1) : 1.0
  end
  
  # <b>tentakel</b> bestimmt die Verwendung der Eigenschaft - @eigens.verwendung nicht benutzt, um Abstimmungsproblemen zu entkommen
  
  def tentakel(check)    
    files = check > 0 ? Feld.find(:all, :conditions => ["eigenschaft = ? AND redundant = ?", check, 0] ) : [0]  # wenigstens die eigene Eigenschaft muss vorliegen, wobei keine Redundanz vorliegen darf
    #files = check > 0 ? Feld.find_all_by_eigenschaft(check) : [0]  # wenigstens die eigene Eigenschaft muss vorliegen
    return  files.length, files.length.to_f/@p
  end  
   
  def distanz(feldname)
    id      = 0
    # Vorlufer und Nachfolger sind bereits geprft
    if    @vor   && feldname == 'vorgaenger'
      id    = @vor.id
      check = @vor.vorgaenger
    elsif @nach  && feldname == 'nachfolger'
      id    = @nach.id
      check = @nach.nachfolger
    end
    if id == 0 || nil
      return 0,0
    elsif id == -1      # Endstation
      return @p+1, (@p+1).to_f/@p
    else
      return weiter(check, feldname, 1, [self.id, id ])
    end
  end

  def weiter(check, feldname, lauf, list)    
    if check > 0
      id      = eval('Feld.find(check).' + feldname)
      if id > 0 && !list.include?(id)
        lauf    += 1
        list    << id
        lauf, x = weiter(id, feldname, lauf, list)
      end
    end    
    return lauf, lauf.to_f/@p
  end
  
  def typ(attr)
    limit = 1.0/@p
    ge    = attr['eigengewicht']   > limit ? '+' : '-'
    gf    = attr['profilgewicht']  > limit ? '+' : '-'
    gp    = attr['portalgewicht']  > limit ? '+' : '-'
    gx    = attr['exitgewicht']    > limit ? '+' : '-'
    return ge+gf, gp+gx    
  end
  
  def related(check, todo)
    return      unless check.to_i > 0
    case todo
    when 'red'
      name  = 'Redundanz'
      feld  = 'redundant'
      inst  = '@master'
    when 'vor'
      name  = 'Vorg&auml;nger'
      feld  = 'vorgaenger'
      inst  = '@vor'
    when 'nach'
      name  = 'Nachfolger'
      feld  = 'nachfolger'
      inst  = '@nach'
    end
    
    if self.id == check.to_i
      @errors.add("#{feld}", "Das Feld \"#{self.name}\" darf sich nicht selbst als #{name} haben") 
    else
      begin
        related = Feld.find(check.to_i)
      rescue
        @errors.add("#{feld}", "Das Feld \"#{check.to_i}\" kann nicht als #{name} dienen, da es nicht gefunden werden konnte. 
                    Dies k&ouml;nnte daran liegen, dass Sie auf einen &uuml;berholten Stand der Anzeige gebl&auml;ttert haben
                    und l&auml;sst sich durch \"Neuladen\" des Browsers bereinigen.") 
      end
    end

    if related && related.redundant > 0    
      begin
        red = Feld.find(related.redundant)
        @errors.add("#{feld}", "Das Feld \"#{related.name}\" kann nicht als #{name} dienen, da es nur ein redundantes Feld ist. 
                Das f&uuml;hrende Feld ist \"#{red.name}\" ")
      rescue
        @errors.add("#{feld}", "Das Feld \"#{related.name}\" kann nicht als #{name} dienen, da es nur ein redundantes Feld ist. 
                Das f&uuml;hrende Feld ist als \"#{related.redundant}\" angegeben, konnte jedoch nicht gefunden werden. ")        
      
      end 
      related = nil
    end
    eval("#{inst} = related")
  end 
  
end
